@@ -15,7 +15,7 @@
 
 There should be a file 'ftesting.zcml' in the current directory.
 
-$Id: functional.py 110210 2010-03-26 17:01:02Z mgedmin $
+$Id: functional.py 90031 2008-08-20 17:07:35Z fdrake $
 """
 import logging
 import os.path
@@ -24,7 +24,6 @@ import rfc822
 import sys
 import traceback
 import unittest
-import doctest
 from StringIO import StringIO
 from Cookie import SimpleCookie
 from transaction import abort, commit
@@ -37,6 +36,7 @@ from zope.publisher.browser import BrowserRequest, setDefaultSkin
 from zope.publisher.http import HTTPRequest
 from zope.publisher.publish import publish
 from zope.security.interfaces import Forbidden, Unauthorized
+from zope.testing import doctest
 
 import zope.app.testing.setup
 from zope.app.appsetup.appsetup import multi_database
@@ -163,7 +163,8 @@ class FunctionalTestSetup(object):
 
     __shared_state = { '_init': False }
 
-    def __init__(self, config_file=None, database_names=None):
+    def __init__(self, config_file=None, database_names=None,
+                 product_config=None):
         """Initializes Zope 3 framework.
 
         Creates a volatile memory storage.  Parses Zope3 configuration files.
@@ -187,16 +188,25 @@ class FunctionalTestSetup(object):
             # Make it silent but keep the log available for debugging
             logging.root.addHandler(logging.StreamHandler(self.log))
 
+            product_configs = []
+            if product_config:
+                configs = zope.app.appsetup.product.loadConfiguration(
+                    StringIO(product_config))
+                zope.app.appsetup.product.setProductConfigurations([
+                    zope.app.appsetup.product.FauxConfiguration(name, values)
+                    for name, values in configs.items()
+                    ])
+
             self._base_storages = {}
             self.db = multi_database(
                 BaseDatabaseFactory(name, self._base_storages)
                 for name in database_names
                 )[0][0]
-            self.dbstack = []
             self.app = Debugger(self.db, config_file)
 
             self.connection = None
             self._config_file = config_file
+            self._product_config = product_config
             self._database_names = database_names
             self._init = True
 
@@ -213,6 +223,10 @@ class FunctionalTestSetup(object):
             raise NotImplementedError('Already configured'
                                       ' with a different config file')
 
+        elif product_config and product_config != self._product_config:
+            raise NotImplementedError('Already configured'
+                                      ' with different product configuration')
+
         elif database_names and database_names != self._database_names:
             # Running different tests with different configurations is not
             # supported at the moment
@@ -234,46 +248,36 @@ class FunctionalTestSetup(object):
     base_storage = property(_get_base_storage, _set_base_storage)
 
     def _close_databases(self):
-        # This is really careful to unregister the databases before attempting
-        # to close anything.  Zope Corporation has a couple of large
-        # multi-database applications that get bitten if we're not careful
-        # like this, but we've not been able to write a concise test case yet.
         base = component.getGlobalSiteManager()
-        dbs = []
-        for name, db in list(component.getUtilitiesFor(IDatabase)):
-            ok = base.unregisterUtility(db, IDatabase, name)
-            assert ok
-            dbs.append(db)
-        if self.connection:
-            self.connection.close()
-            self.connection = None
-        for db in dbs:
+        for name, db in self.db.databases.iteritems():
             db.close()
+            base.unregisterUtility(db, IDatabase, name)
 
     def setUp(self):
         """Prepares for a functional test case."""
         # Tear down the old demo storages (if any) and create fresh ones
         abort()
-        self.dbstack.append((self.db, self.connection))
-        self.connection = None
+        self._close_databases()
         self.db = self.app.db = multi_database(
             DerivedDatabaseFactory(name, self._base_storages)
             for name in self._database_names
             )[0][0]
+        self.connection = None
 
     def tearDown(self):
         """Cleans up after a functional test case."""
         abort()
+        if self.connection:
+            self.connection.close()
+            self.connection = None
         self._close_databases()
-        self.db, self.connection = self.dbstack.pop()
         setSite(None)
 
     def tearDownCompletely(self):
         """Cleans up the setup done by the constructor."""
-        self._close_databases()
-        assert self.dbstack == []
         zope.app.testing.setup.placefulTearDown()
         self._config_file = False
+        self._product_config = None
         self._database_names = None
         self._init = False
 
@@ -295,14 +299,17 @@ class ZCMLLayer:
 
     __bases__ = ()
 
-    def __init__(self, config_file, module, name, allow_teardown=False):
+    def __init__(self, config_file, module, name, allow_teardown=False,
+                 product_config=None):
         self.config_file = config_file
         self.__module__ = module
         self.__name__ = name
         self.allow_teardown = allow_teardown
+        self.product_config = product_config
 
     def setUp(self):
-        self.setup = FunctionalTestSetup(self.config_file)
+        self.setup = FunctionalTestSetup(
+            self.config_file, product_config=self.product_config)
 
     def tearDown(self):
         self.setup.tearDownCompletely()
@@ -321,7 +328,7 @@ def defineLayer(name, zcml='test.zcml', allow_teardown=False):
     """
     globals = sys._getframe(1).f_globals
     globals[name] = ZCMLLayer(
-        os.path.join(os.path.split(globals['__file__'])[0], zcml),
+        os.path.join(os.path.dirname(globals['__file__']), zcml),
         globals['__name__'],
         name,
         allow_teardown=allow_teardown,
